from sklearn import datasets
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler, MinMaxScaler
from sklearn.decomposition import PCA

import numpy

from qiskit import Aer
from qiskit.aqua import QuantumInstance
from qiskit.aqua.algorithms import QSVM
from qiskit.aqua.components.feature_maps import SecondOrderExpansion

# Daten laden und nach Daten und Befunden auftrennen.
daten, befunde = datasets.load_breast_cancer(True)

# Daten in ein Trainingsset (60%) und ein Testset (40%) splitten. Der Split ist
# zufällig, zum Debuggen kann es sinnvoll sein den RNG-Seed zu fixieren, um
# immer die gleichen Ergebnisse zu erhalten:
# train_test_split(daten, befunde, test_size=0.4, random_state=1337)
train_d, test_d, train_b, test_b = train_test_split(daten, befunde, test_size=0.4)

# Standardisieren der Daten
ssc = StandardScaler().fit(train_d)
train_d = ssc.transform(train_d)
test_d = ssc.transform(test_d)

# Hier wird per Hauptkomponentenanalyse die Anzahl der Features auf die Anzahl
# der verwendeten Qubits reduziert.
pca = PCA(n_components = 2).fit(train_d)
train_d = pca.transform(train_d)
test_d = pca.transform(test_d)

# Skalierung der Daten auf das Intervall (-1, +1)
mms = MinMaxScaler((-1, 1)).fit(numpy.append(train_d, test_d, axis=0))
train_d = mms.transform(train_d)
test_d = mms.transform(test_d)

# Die Anzahl der Samples zum Training aus jeder Verteilung wird gewählt, 20
# Trainingssample, 10 Test-Samples.
# 
# Das Indexing nutzt das Boolsche-Array-Indexing von Numpys ndarrays [1], um
# auch die Samples nach Befund zu trennen: train_b (Befunde) wird für jeden
# Index mit "0" bzw. "1" verglichen, und train_d (Daten) damit gefiltert.
# 
# [1] https://docs.scipy.org/doc/numpy/reference/arrays.indexing.html#boolean-array-indexing
training_input = {
  'A': (train_d[train_b==0, :])[:20],
  'B': (train_d[train_b==1, :])[:20]
}

test_input = {
  'A': (test_d[test_b==0, :])[:10],
  'B': (test_d[test_b==1, :])[:10]
}

# Eine Feature-Map wird erstellt und eine SVM damit konstruiert. Hier wird der
# eigentliche Quantenschaltkreis definiert.
feature_map = SecondOrderExpansion(feature_dimension=2, depth=2)
qsvm = QSVM(feature_map, training_input, test_input)

# Auf Wunsch kann Qiskit den generierten Schaltkreis zeichnen, als echte Grafik,
# als ASCII-Art oder als LaTeX-Code [2]. Zum Plotten müssen zwei Punkte
# übergeben werden, aber ihre Wahlt ist für das Layout des Schaltkreises
# unerheblich.
#
# Der Aufruf von decompose() führt zu einer detaillierteren Darstellung des
# Schaltkreises. Er kann auch weggelassen werden (qc.draw(…)) oder mehrfach
# ausgeführt werden (qc.decompose().decompose().draw(…)) für unterschiedliche
# Sichten auf den Schaltkreis.
# 
# [2] https://qiskit.org/documentation/stubs/qiskit.circuit.QuantumCircuit.draw.html#qiskit.circuit.QuantumCircuit.draw
#qc = qsvm.construct_circuit(training_input['A'][0], training_input['A'][1])
#qc.decompose().draw(output='mpl', filename='circuit.png')

# Mit einem Simulator als Backend wird eine QuantumInstance konstruiert und
# ausgeführt. Alternativ könnte auch ein echter Quantencomputer bei IBM als
# Backend genutzt werden [3,4].
# 
# Zum Debuggen kann es wieder sinnvoll sein die Seeds der beteilgten RNGs zu
# fixieren, um immer die gleichen Ergebnisse zu erhalten:
# QuantumInstance(Aer.get_backend('qasm_simulator'), shots=1024,
#     seed_simulator=4711, seed_transpiler=4711)
# 
# [3] https://qiskit.org/documentation/install.html#access-ibm-quantum-systems
# [4] https://qiskit.org/documentation/apidoc/ibmq_provider.html
instanz = QuantumInstance(Aer.get_backend('qasm_simulator'), shots=1024)
erg = qsvm.run(instanz)

# Anzeigen des Results, wie gut die QSVM klassifiziert hat
print("Genauigkeit:", erg['testing_accuracy'])
